from pygame import mixer
import pygame

from spacevader.entities.constant import Constant

class Bullet:

    def __init__(self, screen=None):
        self._icon = pygame.image.load(Constant.BULLET_ICON)
        self.x = Constant.BULLET_X_INITIAL
        self.y = Constant.BULLET_Y_INITIAL
        self._x_change = Constant.BULLET_X_CHANGE
        self._y_change = Constant.BULLET_Y_CHANGE
        self.state = 'READY'
        self._screen = screen

    def fire(self, sound=False):
        if sound:
            bullet_sound = mixer.Sound(Constant.BULLET_SOUND)
            bullet_sound.play()
        self.state = 'FIRE'
        self._screen.blit(self._icon, (self.x + 16, self.y + 10))


from itertools import count
import math
import random

from pygame import mixer
import pygame

from spacevader.entities.constant import Constant

class Alien:

    _ids = count(0)

    def __init__(self, x=None, y=None, x_change=None, y_change=None, screen=None):
        self.id = next(self._ids)
        self._icon = pygame.image.load(Constant.ALIEN_ICON)
        self._x = x
        self._y = y
        self.x_change = x_change
        self._y_change = y_change
        self._screen = screen
        self.invaded = False

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        if x <= Constant.LEFT_BOUNDARY:
            self.x_change = Constant.ALIEN_X_MOVEMENT_DELTA
            self.y += self._y_change
        elif x >= Constant.RIGHT_BOUNDARY:
            self.x_change = Constant.ALIEN_X_MOVEMENT_DELTA * -1
            self.y += self._y_change
        self._x = x

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, y):
        self._y = y
        if self._y > Constant.INVADE_BOUNDARY:
            self.invaded = True

    @classmethod
    def game_over(cls, screen):
        game_over_font = pygame.font.Font(Constant.FONT, Constant.GAME_OVER_FONT_SIZE)
        game_over_text = game_over_font.render('GAME OVER', True, (255, 255, 255))
        screen.blit(game_over_text, (Constant.GAME_OVER_X, Constant.GAME_OVER_Y))

        restart_icon = pygame.image.load(Constant.RESTART_ICON)
        screen.blit(restart_icon, (Constant.RESTART_X, Constant.RESTART_Y))

    def has_collided(self, bullet):
        distance = math.sqrt(math.pow(self._x - bullet.x, 2) + math.pow(self._y - bullet.y, 2))
        return distance < Constant.COLLISION_DELTA

    def explode(self, bullet):
        explosion_sound = mixer.Sound(Constant.EXPLOSION_SOUND)
        explosion_sound.play()
        # TODO: Check if this is necessary for NORMAL mode
        self.x = random.randint(Constant.LEFT_BOUNDARY, Constant.RIGHT_BOUNDARY)
        self.y = random.randint(Constant.ALIEN_SPAWN_TOP_BOUNDARY, Constant.ALIEN_SPAWN_BOTTOM_BOUNDARY)
        bullet.y = Constant.BULLET_Y_INITIAL
        bullet.state = 'READY'

    def render(self):
        self._screen.blit(self._icon, (self._x, self._y))

    @classmethod
    def dispose(cls, aliens):
        for alien in aliens:
            alien.y = 2000



import pygame

from spacevader.entities.constant import Constant

# pygame.init()

class Player:

    def __init__(self, screen=None):
        # TODO: Evaluate use of meta class
        self._icon = pygame.image.load(Constant.PLAYER_ICON)
        self._x = Constant.PLAYER_X_INITIAL
        self.y = Constant.PLAYER_Y_INITIAL
        self.x_change = Constant.PLAYER_X_CHANGE
        self._screen = screen

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        if x <= Constant.LEFT_BOUNDARY:
            self._x = Constant.LEFT_BOUNDARY
        elif x >= Constant.RIGHT_BOUNDARY:
            self._x = Constant.RIGHT_BOUNDARY
        else:
            self._x = x

    def delta(self, event_type, event_key):
        movement_switcher = {
            pygame.KEYDOWN: {
                pygame.K_LEFT: Constant.PLAYER_X_MOVEMENT_DELTA * -1,
                pygame.K_RIGHT: Constant.PLAYER_X_MOVEMENT_DELTA
            },
            pygame.KEYUP: {
                pygame.K_LEFT: Constant.PLAYER_X_MOVEMENT_DELTA * 0,
                pygame.K_RIGHT: Constant.PLAYER_X_MOVEMENT_DELTA * 0
            }
        }
        return movement_switcher[event_type].get(event_key)

    def render(self):
        self._screen.blit(self._icon, (self.x, self.y))


import random
import math
import os
import sys

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = 'True'

from pygame import mixer
import pygame
import psutil

from spacevader.entities.constant import Constant
from spacevader.entities.player import Player
from spacevader.entities.alien import Alien
from spacevader.entities.bullet import Bullet

current_dir = os.path.dirname(os.path.realpath(__file__))

# Initialize pygame
pygame.init()

# Create a screen
screen = pygame.display.set_mode((800, 600))

# Set background
# TODO: Set dynamic background
background = pygame.image.load(Constant.BACKGROUND_ICON)

# Load background sound
mixer.music.load(Constant.BACKGROUND_SOUND)
mixer.music.play(-1)

# Set the score
score = 0
font = pygame.font.Font(Constant.FONT, 32)

# Set title and icon
pygame.display.set_caption(Constant.GAME_NAME)
icon = pygame.image.load(Constant.GAME_ICON)
pygame.display.set_icon(icon)

# Player
player = Player(screen)

# Alien
aliens = []

# Bullet
bullet = Bullet(screen)

def show_score():
    score_value = font.render(f'{score}', True, (255, 255, 255))
    screen.blit(score_value, (Constant.SCORE_BOARD_X, Constant.SCORE_BOARD_Y))

def run():
    global player
    global aliens
    global bullet
    global score

    running = True
    while running:
        
        # Fill screen with black
        screen.fill((0, 0, 0))

        # Persist background image
        screen.blit(background, (0, 0))
        
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False

            if event.type == pygame.KEYDOWN:
                if event.key in (pygame.K_LEFT, pygame.K_RIGHT):
                    player.x_change = player.delta(event.type, event.key)
                if event.key == pygame.K_SPACE and bullet.state == 'READY':
                    bullet.x = player.x
                    bullet.fire(sound=True)

            if event.type == pygame.KEYUP:
                if event.key in (pygame.K_LEFT, pygame.K_RIGHT):
                    player.x_change = player.delta(event.type, event.key)

            if event.type == pygame.MOUSEBUTTONUP:
                mouse = pygame.mouse.get_pos()
                if Constant.RESTART_X + Constant.GAME_OVER_FONT_SIZE > mouse[0] > Constant.RESTART_X and Constant.RESTART_Y + Constant.GAME_OVER_FONT_SIZE > mouse[1] > Constant.RESTART_Y:
                    try:
                        proc = psutil.Process(os.getpid())
                        for handler in proc.get_open_files() + proc.connections():
                            os.close(handler.fd)
                    except Exception as e:
                        print(e)

                    python = sys.executable
                    os.execl(python, python, *sys.argv)

        player.x += player.x_change

        for i in range(score + 1):

            alien = Alien(x=random.randint(0, 736), y=random.randint(50, 150), x_change=random.choice([-5, 5]), y_change=40, screen=screen)
            aliens.append(alien)

            aliens[i].x += aliens[i].x_change

            # Collision
            collision = aliens[i].has_collided(bullet)
            if collision:
                aliens[i].explode(bullet)
                score += 1

            # Invasion
            if aliens[i].invaded:
                Alien.dispose(aliens)
                Alien.game_over(screen)
                break

            aliens[i].render()
            

        # Bullet movement
        if bullet.y <= 0:
            bullet.y = 480
            bullet.state = 'READY'
        if bullet.state == 'FIRE':
            bullet.fire()
            bullet.y -= bullet._y_change

        player.render()
        show_score()
        pygame.display.update()

if __name__ == '__main__':
    run()


import os

class Constant:

    CURR_DIR = os.path.dirname(os.path.realpath(__file__))

    GAME_NAME = 'Spacevader'
    GAME_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/ufo.png')

    LEFT_BOUNDARY = 0
    RIGHT_BOUNDARY = 736
    INVADE_BOUNDARY = 440

    BACKGROUND_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/background-2.jpg')
    BACKGROUND_SOUND = os.path.join(f'{CURR_DIR}/../', 'sounds/background.wav')

    FONT = 'freesansbold.ttf'

    PLAYER_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/player.png')
    PLAYER_X_INITIAL = 370
    PLAYER_Y_INITIAL = 480
    PLAYER_X_CHANGE = 0
    PLAYER_X_MOVEMENT_DELTA = 7

    ALIEN_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/alien.png')
    ALIEN_X_MOVEMENT_DELTA = 4
    ALIEN_SPAWN_TOP_BOUNDARY = 50
    ALIEN_SPAWN_BOTTOM_BOUNDARY = 150

    EXPLOSION_SOUND = os.path.join(f'{CURR_DIR}/../', 'sounds/explosion.wav')

    COLLISION_DELTA = 27

    GAME_OVER_FONT_SIZE = 64
    GAME_OVER_X = 200
    GAME_OVER_Y = 250

    RESTART_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/restart.png')
    RESTART_X = 370
    RESTART_Y = GAME_OVER_Y + GAME_OVER_FONT_SIZE + 15
    
    BULLET_ICON = os.path.join(f'{CURR_DIR}/../', 'icons/bullet.png')
    BULLET_X_INITIAL = 0
    BULLET_Y_INITIAL = 480
    BULLET_X_CHANGE = 0
    BULLET_Y_CHANGE = 20
    BULLET_SOUND = os.path.join(f'{CURR_DIR}/../', 'sounds/laser.wav')

    SCORE_BOARD_X = 750
    SCORE_BOARD_Y = 10
    